import 'package:flutter/gestures.dart';
import 'package:flutter/widgets.dart';

import 'flutter/widgets/scrollable.dart';
import 'overscroll_gestures.dart';

class OverscrollScrollable extends FlutterScrollable {
  const OverscrollScrollable({
    super.key,
    super.axisDirection,
    super.controller,
    super.physics,
    required super.viewportBuilder,
    super.incrementCalculator,
    super.excludeFromSemantics,
    super.semanticChildCount,
    super.dragStartBehavior,
    super.restorationId,
    super.scrollBehavior,
    super.clipBehavior,
  });

  static FlutterScrollIncrementCalculator? _getFlutterScrollIncrementCalculator(
    Scrollable scrollable,
  ) {
    final incrementCalculator = scrollable.incrementCalculator;
    if (incrementCalculator == null) {
      return null;
    }
    return (details) {
      return incrementCalculator(
        ScrollIncrementDetails(
          metrics: details.metrics,
          type: ScrollIncrementType.values[details.type.index],
        ),
      );
    };
  }

  /// 将 Scrollable 转换为 OverscrollScrollable
  OverscrollScrollable.from(Scrollable scrollable)
      : super(
          key: scrollable.key,
          axisDirection: scrollable.axisDirection,
          controller: scrollable.controller,
          physics: scrollable.physics,
          viewportBuilder: scrollable.viewportBuilder,
          incrementCalculator: _getFlutterScrollIncrementCalculator(scrollable),
          excludeFromSemantics: scrollable.excludeFromSemantics,
          semanticChildCount: scrollable.semanticChildCount,
          dragStartBehavior: scrollable.dragStartBehavior,
          restorationId: scrollable.restorationId,
          scrollBehavior: scrollable.scrollBehavior,
          clipBehavior: scrollable.clipBehavior,
        );

  @override
  FlutterScrollableState createState() => _OverscrollScrollableState();
}

class _OverscrollScrollableState extends FlutterScrollableState {
  /// 是否已经滚动到边界
  bool _overscroll = false;

  /// 代理手势更新事件
  GestureDragUpdateCallback? _proxyOnUpdate(GestureDragUpdateCallback? parent) {
    // 如果已经滚动到了边界则不再响应移动事件
    return _overscroll ? null : parent;
  }

  /// 处理手势移动事件
  _onMoveEvent(PointerMoveEvent event, Offset delta, double primaryDelta) {
    // 如果已经滚动到边界并且又继续移动
    if (_overscroll && primaryDelta != 0.0) {
      // 边界滚动事件通知
      OverscrollNotification(
        metrics: position.copyWith(),
        context: context,
        overscroll: primaryDelta,
        dragDetails: DragUpdateDetails(
          sourceTimeStamp: event.timeStamp,
          delta: delta,
          primaryDelta: primaryDelta,
          globalPosition: event.position,
          localPosition: event.localPosition,
        ),
      ).dispatch(context);
    }
  }

  @override
  set gestureRecognizers(value) {
    super.gestureRecognizers = value.map((key, value) {
      if (key == VerticalDragGestureRecognizer) {
        return MapEntry(
          OverscrollVerticalDragGestureRecognizer,
          GestureRecognizerFactoryWithHandlers<
              OverscrollVerticalDragGestureRecognizer>(
            () => OverscrollVerticalDragGestureRecognizer(
              supportedDevices: configuration.dragDevices,
              proxyOnUpdate: _proxyOnUpdate,
              onPointerMoveEvent: _onMoveEvent,
            ),
            (instance) => value.initializer(instance),
          ),
        );
      } else if (key == HorizontalDragGestureRecognizer) {
        return MapEntry(
          OverscrollHorizontalDragGestureRecognizer,
          GestureRecognizerFactoryWithHandlers<
              OverscrollHorizontalDragGestureRecognizer>(
            () => OverscrollHorizontalDragGestureRecognizer(
              supportedDevices: configuration.dragDevices,
              proxyOnUpdate: _proxyOnUpdate,
              onPointerMoveEvent: _onMoveEvent,
            ),
            (instance) => value.initializer(instance),
          ),
        );
      }
      return MapEntry(key, value);
    });
  }

  @override
  Widget build(BuildContext context) {
    return NotificationListener<ScrollNotification>(
      onNotification: (ScrollNotification notification) {
        if (notification is OverscrollNotification) {
          // 设置边界滚动状态
          _overscroll = true;
        } else if (notification is ScrollEndNotification) {
          // 重置边界滚动状态
          _overscroll = false;
        }
        return false;
      },
      child: super.build(context),
    );
  }
}
